@@ -52,11 +52,13 @@
 #include <sensor_msgs/NavSatFix.h>
 #include <sensor_msgs/TimeReference.h>
 // GNSS msg includes
+#include <septentrio_gnss_driver/AIMPlusStatus.h>
 #include <septentrio_gnss_driver/AttCovEuler.h>
 #include <septentrio_gnss_driver/AttEuler.h>
 #include <septentrio_gnss_driver/BaseVectorCart.h>
 #include <septentrio_gnss_driver/BaseVectorGeod.h>
 #include <septentrio_gnss_driver/BlockHeader.h>
+#include <septentrio_gnss_driver/GALAuthStatus.h>
 #include <septentrio_gnss_driver/MeasEpoch.h>
 #include <septentrio_gnss_driver/MeasEpochChannelType1.h>
 #include <septentrio_gnss_driver/MeasEpochChannelType2.h>
@@ -64,6 +66,8 @@
 #include <septentrio_gnss_driver/PVTGeodetic.h>
 #include <septentrio_gnss_driver/PosCovCartesian.h>
 #include <septentrio_gnss_driver/PosCovGeodetic.h>
+#include <septentrio_gnss_driver/RFBand.h>
+#include <septentrio_gnss_driver/RFStatus.h>
 #include <septentrio_gnss_driver/ReceiverTime.h>
 #include <septentrio_gnss_driver/VectorInfoCart.h>
 #include <septentrio_gnss_driver/VectorInfoGeod.h>
@@ -81,8 +85,8 @@
 #include <septentrio_gnss_driver/INSNavGeod.h>
 #include <septentrio_gnss_driver/VelSensorSetup.h>
 // Rosaic includes
-#include <septentrio_gnss_driver/communication/settings.h>
-#include <septentrio_gnss_driver/parsers/string_utilities.h>
+#include <septentrio_gnss_driver/communication/settings.hpp>
+#include <septentrio_gnss_driver/parsers/string_utilities.hpp>
 
 // Timestamp in nanoseconds (Unix epoch)
 typedef uint64_t Timestamp;
@@ -96,18 +100,22 @@ typedef geometry_msgs::Quaternion QuaternionMsg;
 typedef geometry_msgs::PoseWithCovarianceStamped PoseWithCovarianceStampedMsg;
 typedef geometry_msgs::TwistWithCovarianceStamped TwistWithCovarianceStampedMsg;
 typedef geometry_msgs::TransformStamped TransformStampedMsg;
-typedef gps_common::GPSFix GPSFixMsg;
-typedef gps_common::GPSStatus GPSStatusMsg;
+typedef gps_common::GPSFix GpsFixMsg;
+typedef gps_common::GPSStatus GpsStatusMsg;
 typedef sensor_msgs::NavSatFix NavSatFixMsg;
 typedef sensor_msgs::NavSatStatus NavSatStatusMsg;
 typedef sensor_msgs::TimeReference TimeReferenceMsg;
 typedef sensor_msgs::Imu ImuMsg;
-typedef nav_msgs::Odometry LocalizationUtmMsg;
+typedef nav_msgs::Odometry LocalizationMsg;
 
 // Septentrio GNSS SBF messages
+typedef septentrio_gnss_driver::AIMPlusStatus AimPlusStatusMsg;
 typedef septentrio_gnss_driver::BaseVectorCart BaseVectorCartMsg;
 typedef septentrio_gnss_driver::BaseVectorGeod BaseVectorGeodMsg;
 typedef septentrio_gnss_driver::BlockHeader BlockHeaderMsg;
+typedef septentrio_gnss_driver::GALAuthStatus GalAuthStatusMsg;
+typedef septentrio_gnss_driver::RFStatus RfStatusMsg;
+typedef septentrio_gnss_driver::RFBand RfBandMsg;
 typedef septentrio_gnss_driver::MeasEpoch MeasEpochMsg;
 typedef septentrio_gnss_driver::MeasEpochChannelType1 MeasEpochChannelType1Msg;
 typedef septentrio_gnss_driver::MeasEpochChannelType2 MeasEpochChannelType2Msg;
@@ -158,14 +166,16 @@ inline Timestamp timestampFromRos(const TimestampRos& tsr) { return tsr.toNSec()
 /**
  * @brief Log level for ROS logging
  */
-enum LogLevel
-{
-    DEBUG,
-    INFO,
-    WARN,
-    ERROR,
-    FATAL
-};
+namespace log_level {
+    enum LogLevel
+    {
+        DEBUG,
+        INFO,
+        WARN,
+        ERROR,
+        FATAL
+    };
+} // namespace log_level
 
 /**
  * @class ROSaicNodeBase
@@ -174,30 +184,45 @@ enum LogLevel
 class ROSaicNodeBase
 {
 public:
-    ROSaicNodeBase() : pNh_(new ros::NodeHandle("~")), tfListener_(tfBuffer_) {}
+    ROSaicNodeBase() :
+        pNh_(new ros::NodeHandle("~")), tfListener_(tfBuffer_), lastTfStamp_(0)
+    {
+    }
 
     virtual ~ROSaicNodeBase() {}
 
+    const Settings* settings() const { return &settings_; }
+
     void registerSubscriber()
     {
-        ros::NodeHandle nh;
-        if (settings_.ins_vsm_ros_source == "odometry")
-            odometrySubscriber_ = nh.subscribe<nav_msgs::Odometry>(
-                "odometry_vsm", 10, &ROSaicNodeBase::callbackOdometry, this);
-        else if (settings_.ins_vsm_ros_source == "twist")
-            twistSubscriber_ = nh.subscribe<TwistWithCovarianceStampedMsg>(
-                "twist_vsm", 10, &ROSaicNodeBase::callbackTwist, this);
+        try
+        {
+            ros::NodeHandle nh;
+            if (settings_.ins_vsm_ros_source == "odometry")
+                odometrySubscriber_ = nh.subscribe<nav_msgs::Odometry>(
+                    "odometry_vsm", 10, &ROSaicNodeBase::callbackOdometry, this);
+            else if (settings_.ins_vsm_ros_source == "twist")
+                twistSubscriber_ = nh.subscribe<TwistWithCovarianceStampedMsg>(
+                    "twist_vsm", 10, &ROSaicNodeBase::callbackTwist, this);
+        } catch (const std::runtime_error& ex)
+        {
+            this->log(log_level::ERROR, "Subscriber initialization failed due to: " +
+                                            std::string(ex.what()) + ".");
+        }
     }
 
     /**
-     * @brief Gets an integer or unsigned integer value from the parameter server
-     * @param[in] name The key to be used in the parameter server's dictionary
-     * @param[out] val Storage for the retrieved value, of type U, which can be
-     * either unsigned int or int
-     * @param[in] defaultVal Value to use if the server doesn't contain this
-     * parameter
+     * @brief Gets an integer or unsigned integer value from the
+     * parameter server
+     * @param[in] name The key to be used in the parameter server's
+     * dictionary
+     * @param[out] val Storage for the retrieved value, of type U, which
+     * can be either unsigned int or int
+     * @param[in] defaultVal Value to use if the server doesn't contain
+     * this parameter
      */
-    bool getUint32Param(const std::string& name, uint32_t& val, uint32_t defaultVal)
+    bool getUint32Param(const std::string& name, uint32_t& val,
+                        uint32_t defaultVal) const
     {
         int32_t tempVal;
         if ((!pNh_->getParam(name, tempVal)) || (tempVal < 0))
@@ -211,14 +236,15 @@ public:
 
     /**
      * @brief Gets parameter of type T from the parameter server
-     * @param[in] name The key to be used in the parameter server's dictionary
+     * @param[in] name The key to be used in the parameter server's
+     * dictionary
      * @param[out] val Storage for the retrieved value, of type T
-     * @param[in] defaultVal Value to use if the server doesn't contain this
-     * parameter
+     * @param[in] defaultVal Value to use if the server doesn't contain
+     * this parameter
      * @return True if it could be retrieved, false if not
      */
     template <typename T>
-    bool param(const std::string& name, T& val, const T& defaultVal)
+    bool param(const std::string& name, T& val, const T& defaultVal) const
     {
         return pNh_->param(name, val, defaultVal);
     };
@@ -228,23 +254,23 @@ public:
      * @param[in] logLevel Log level
      * @param[in] s String to log
      */
-    void log(LogLevel logLevel, const std::string& s)
+    void log(log_level::LogLevel logLevel, const std::string& s) const
     {
         switch (logLevel)
         {
-        case LogLevel::DEBUG:
+        case log_level::DEBUG:
             ROS_DEBUG_STREAM(ros::this_node::getName() << ": " << s);
             break;
-        case LogLevel::INFO:
+        case log_level::INFO:
             ROS_INFO_STREAM(ros::this_node::getName() << ": " << s);
             break;
-        case LogLevel::WARN:
+        case log_level::WARN:
             ROS_WARN_STREAM(ros::this_node::getName() << ": " << s);
             break;
-        case LogLevel::ERROR:
+        case log_level::ERROR:
             ROS_ERROR_STREAM(ros::this_node::getName() << ": " << s);
             break;
-        case LogLevel::FATAL:
+        case log_level::FATAL:
             ROS_FATAL_STREAM(ros::this_node::getName() << ": " << s);
             break;
         default:
@@ -256,7 +282,7 @@ public:
      * @brief Gets current timestamp
      * @return Timestamp
      */
-    Timestamp getTime() { return ros::Time::now().toNSec(); }
+    Timestamp getTime() const { return ros::Time::now().toNSec(); }
 
     /**
      * @brief Publishing function
@@ -282,7 +308,7 @@ public:
      * @brief Publishing function for tf
      * @param[in] msg ROS localization message to be converted to tf
      */
-    void publishTf(const LocalizationUtmMsg& loc)
+    void publishTf(const LocalizationMsg& loc)
     {
         if (std::isnan(loc.pose.pose.orientation.w))
             return;
@@ -347,6 +373,36 @@ public:
         tf2Publisher_.sendTransform(transformStamped);
     }
 
+    /**
+     * @brief Set INS to true
+     */
+    void setIsIns() { capabilities_.is_ins = true; }
+
+    /**
+     * @brief Set has heading to true
+     */
+    void setHasHeading() { capabilities_.has_heading = true; }
+
+    /**
+     * @brief Set improved VSM handling to true
+     */
+    void setImprovedVsmHandling() { capabilities_.has_improved_vsm_handling = true; }
+
+    /**
+     * @brief Check if Rx is INS
+     */
+    bool isIns() { return capabilities_.is_ins; }
+
+    /**
+     * @brief Check if Rx has heading
+     */
+    bool hasHeading() { return capabilities_.has_heading; }
+
+    /**
+     * @brief Check if Rx has improved VSM handling
+     */
+    bool hasImprovedVsmHandling() { return capabilities_.has_improved_vsm_handling; }
+
 private:
     void callbackOdometry(const nav_msgs::Odometry::ConstPtr& odo)
     {
@@ -365,93 +421,121 @@ private:
     void processTwist(Timestamp stamp,
                       const geometry_msgs::TwistWithCovariance& twist)
     {
-        time_t epochSeconds = stamp / 1000000000;
-        struct tm* tm_temp = std::gmtime(&epochSeconds);
-        std::stringstream timeUtc;
-        timeUtc << std::setfill('0') << std::setw(2)
-                << std::to_string(tm_temp->tm_hour) << std::setw(2)
-                << std::to_string(tm_temp->tm_min) << std::setw(2)
-                << std::to_string(tm_temp->tm_sec) << "." << std::setw(3)
-                << std::to_string((stamp - (stamp / 1000000000) * 1000000000) /
-                                  1000000);
+        // in case stamp was not set
+        if (stamp == 0)
+            stamp = getTime();
 
-        std::string v_x;
-        std::string v_y;
-        std::string v_z;
-        std::string std_x;
-        std::string std_y;
-        std::string std_z;
-        if (settings_.ins_vsm_ros_config[0])
+        static Eigen::Vector3d vel = Eigen::Vector3d::Zero();
+        static Eigen::Vector3d var = Eigen::Vector3d::Zero();
+        static uint64_t ctr = 0;
+        static Timestamp lastStamp = 0;
+
+        ++ctr;
+        vel[0] += twist.twist.linear.x;
+        vel[1] += twist.twist.linear.y;
+        vel[2] += twist.twist.linear.z;
+        var[0] += twist.covariance[0];
+        var[1] += twist.covariance[7];
+        var[2] += twist.covariance[14];
+
+        // Rx expects averaged velocity at a rate of 2 Hz
+        if ((stamp - lastStamp) >= 495000000) // allow for 5 ms jitter
         {
-            v_x = string_utilities::trimDecimalPlaces(twist.twist.linear.x);
-            if (settings_.ins_vsm_ros_variances_by_parameter)
-                std_x = string_utilities::trimDecimalPlaces(
-                    settings_.ins_vsm_ros_variances[0]);
-            else if (twist.covariance[0] > 0.0)
-                std_x = string_utilities::trimDecimalPlaces(
-                    std::sqrt(twist.covariance[0]));
-            else
+            vel /= ctr;
+            var /= ctr;
+            time_t epochSeconds = stamp / 1000000000;
+            struct tm* tm_temp = std::gmtime(&epochSeconds);
+            std::stringstream timeUtc;
+            timeUtc << std::setfill('0') << std::setw(2)
+                    << std::to_string(tm_temp->tm_hour) << std::setw(2)
+                    << std::to_string(tm_temp->tm_min) << std::setw(2)
+                    << std::to_string(tm_temp->tm_sec) << "." << std::setw(3)
+                    << std::to_string((stamp - (stamp / 1000000000) * 1000000000) /
+                                      1000000);
+
+            std::string v_x;
+            std::string v_y;
+            std::string v_z;
+            std::string std_x;
+            std::string std_y;
+            std::string std_z;
+            if (settings_.ins_vsm_ros_config[0])
             {
-                ROS_ERROR_STREAM("Invalid covariance value for v_x: "
-                                 << std::to_string(twist.covariance[0])
-                                 << ". Ignoring measurement.");
-            }
-        } else
-            std_x = std::to_string(1000000.0);
-        if (settings_.ins_vsm_ros_config[1])
-        {
-            if (settings_.use_ros_axis_orientation)
-                v_y = "-";
-            v_y += string_utilities::trimDecimalPlaces(twist.twist.linear.y);
-            if (settings_.ins_vsm_ros_variances_by_parameter)
-                std_y = string_utilities::trimDecimalPlaces(
-                    settings_.ins_vsm_ros_variances[1]);
-            else if (twist.covariance[7] > 0.0)
-                std_y = string_utilities::trimDecimalPlaces(
-                    std::sqrt(twist.covariance[7]));
-            else
+                v_x = string_utilities::trimDecimalPlaces(vel[0]);
+                if (settings_.ins_vsm_ros_variances_by_parameter)
+                    std_x = string_utilities::trimDecimalPlaces(
+                        settings_.ins_vsm_ros_variances[0]);
+                else if (var[0] > 0.0)
+                    std_x = string_utilities::trimDecimalPlaces(std::sqrt(var[0]));
+                else if (!capabilities_.has_improved_vsm_handling)
+                {
+                    log(log_level::ERROR, "Invalid covariance value for v_x: " +
+                                              std::to_string(var[0]) +
+                                              ". Ignoring measurement.");
+                    v_x = "";
+                    std_x = string_utilities::trimDecimalPlaces(1000000.0);
+                }
+            } else
+                std_x = std::to_string(1000000.0);
+            if (settings_.ins_vsm_ros_config[1])
             {
-                ROS_ERROR_STREAM("Invalid covariance value for v_y: "
-                                 << std::to_string(twist.covariance[1])
-                                 << ". Ignoring measurement.");
-                v_y = "";
+                if (settings_.use_ros_axis_orientation)
+                    v_y = "-";
+                v_y += string_utilities::trimDecimalPlaces(vel[1]);
+                if (settings_.ins_vsm_ros_variances_by_parameter)
+                    std_y = string_utilities::trimDecimalPlaces(
+                        settings_.ins_vsm_ros_variances[1]);
+                else if (var[1] > 0.0)
+                    std_y = string_utilities::trimDecimalPlaces(std::sqrt(var[1]));
+                else if (!capabilities_.has_improved_vsm_handling)
+                {
+                    log(log_level::ERROR, "Invalid covariance value for v_y: " +
+                                              std::to_string(var[1]) +
+                                              ". Ignoring measurement.");
+                    v_y = "";
+                    std_y = string_utilities::trimDecimalPlaces(1000000.0);
+                }
+            } else
                 std_y = string_utilities::trimDecimalPlaces(1000000.0);
-            }
-        } else
-            std_y = string_utilities::trimDecimalPlaces(1000000.0);
-        if (settings_.ins_vsm_ros_config[2])
-        {
-            if (settings_.use_ros_axis_orientation)
-                v_z = "-";
-            v_z += string_utilities::trimDecimalPlaces(twist.twist.linear.z);
-            if (settings_.ins_vsm_ros_variances_by_parameter)
-                std_z = string_utilities::trimDecimalPlaces(
-                    settings_.ins_vsm_ros_variances[2]);
-            else if (twist.covariance[14] > 0.0)
-                std_z = string_utilities::trimDecimalPlaces(
-                    std::sqrt(twist.covariance[14]));
-            else
+            if (settings_.ins_vsm_ros_config[2])
             {
-                ROS_ERROR_STREAM("Invalid covariance value for v_z: "
-                                 << std::to_string(twist.covariance[2])
-                                 << ". Ignoring measurement.");
-                v_z = "";
+                if (settings_.use_ros_axis_orientation)
+                    v_z = "-";
+                v_z += string_utilities::trimDecimalPlaces(vel[2]);
+                if (settings_.ins_vsm_ros_variances_by_parameter)
+                    std_z = string_utilities::trimDecimalPlaces(
+                        settings_.ins_vsm_ros_variances[2]);
+                else if (var[2] > 0.0)
+                    std_z = string_utilities::trimDecimalPlaces(std::sqrt(var[2]));
+                else if (!capabilities_.has_improved_vsm_handling)
+                {
+                    log(log_level::ERROR, "Invalid covariance value for v_z: " +
+                                              std::to_string(var[2]) +
+                                              ". Ignoring measurement.");
+                    v_z = "";
+                    std_z = string_utilities::trimDecimalPlaces(1000000.0);
+                }
+            } else
                 std_z = string_utilities::trimDecimalPlaces(1000000.0);
-            }
-        } else
-            std_z = string_utilities::trimDecimalPlaces(1000000.0);
 
-        std::string velNmea = "$PSSN,VSM," + timeUtc.str() + "," + v_x + "," + v_y +
-                              "," + std_x + "," + std_y + "," + v_z + "," + std_z;
+            std::string velNmea = "$PSSN,VSM," + timeUtc.str() + "," + v_x + "," +
+                                  v_y + "," + std_x + "," + std_y + "," + v_z + "," +
+                                  std_z;
 
-        char crc = std::accumulate(velNmea.begin() + 1, velNmea.end(), 0,
-                                   [](char sum, char ch) { return sum ^ ch; });
+            char crc = std::accumulate(velNmea.begin() + 1, velNmea.end(), 0,
+                                       [](char sum, char ch) { return sum ^ ch; });
 
-        std::stringstream crcss;
-        crcss << std::hex << static_cast<int32_t>(crc);
+            std::stringstream crcss;
+            crcss << std::hex << static_cast<int32_t>(crc);
 
-        velNmea += "*" + crcss.str() + "\r\n";
-        sendVelocity(velNmea);
+            velNmea += "*" + crcss.str() + "\r\n";
+            sendVelocity(velNmea);
+
+            vel = Eigen::Vector3d::Zero();
+            var = Eigen::Vector3d::Zero();
+            ctr = 0;
+            lastStamp = stamp;
+        }
     }
 
 protected:
@@ -479,4 +563,6 @@ private:
     tf2_ros::Buffer tfBuffer_;
     // tf listener
     tf2_ros::TransformListener tfListener_;
+    // Capabilities of Rx
+    Capabilities capabilities_;
 };
\ No newline at end of file
